#include "hime.h"

#include "pho-kbm-name.h"

struct PHO_KBM_NAME kbm_sel[] = {
    {N_ ("標準(零壹)"), "zo"},
    {N_ ("Eten"), "et"},
    {N_ ("Eten, 26 keys"), "et26"},
    {N_ ("Mr. Hsu's"), "hsu"},
    {N_ ("Pinyin"), "pinyin"},
    {N_ ("Pinyin, toneless"), "pinyin-no-tone"},
    {N_ ("Dvorak"), "dvorak"},
    {N_ ("IBM"), "ibm"},
    {N_ ("Mitac"), "mitac"},
    {N_ ("Colemak"), "colemak"},
    {NULL, NULL}};
